/*************************************************************************

  graphics16.c v2.0

    Graphics functions to work in 16 bits for the Game Park GP32.
    graphics16.c is part of the Oankali's GP32 run-time library.

  Copyright (c) 2004,2005 by
  Guillem Reinoso <oankali@nekanium.com>
  www.nekanium.com

  These software is freeware and is provided as is.
  If it doesn't work for you or it makes your PC or GP32 melt down, it's not my
  responsibility.
  If you use it in your programs, cite me please.

  History:
    v1.0
      First public release
    v2.0
      Removed   AddColors(), AddPixel(), AddSmoothedColors(), FillOverlay16(),
                SubPixel(), SubstractColors(), TransBltColored16()
      Added     AddColor16(), BitBlt16() and ReleaseScreenRegion().
                SaveBackScreenRegion(), SubtractColor16(), TransBlt16(),
                TransFade16() and TransFadeMask16().
      Renamed   BitBlt8To16() to BitBltGrayTo16().
      Changed   BitTile16() -> added more paremeters.
      Changed   TransTile16() -> added more paremeters.
      Added     alpha blending lookup table to optimize alpha blending.
                -> InitializeAlphaLUT(), ReleaseAlphaLUT(),
                   gtAlphaLUT, giAlphaLevels and ALT_* constants.
      Optimized BitBltGrayTo16() and OverlayColor16().

**************************************************************************/

#include <gpdef.h>
#include <gpstdlib.h>
#include <gpgraphic.h>

#include "global.h"
#include "graphics16.h"


//*** Public global variables
unsigned char (*gtAlphaLUT)[32][32];  // alpha blending lookup table
int giAlphaLevels;                    // precision of the alpha LUT


//*** Public implementations
void AddColor16(GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height, int color)
{
  int i, j;
  unsigned short *screen;
  int red, green, blue;

  // Adjust to rotated screen
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  i = width;
  width = height;
  height = i;


  // Add color
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  for (j=0; j < height; j++)
  {
    for (i=0; i < width; i++, screen++)
    {
      red   = GetRGBRed  (*screen) + GetRGBRed  (color);
      green = GetRGBGreen(*screen) + GetRGBGreen(color);
      blue  = GetRGBBlue (*screen) + GetRGBBlue (color);
      if (red > 31)   red = 31;
      if (green > 31) green = 31;
      if (blue > 31)  blue = 31;

      *screen = RGB(red, green, blue);
    }

    screen += GPC_LCD_PHYSICAL_W - width;
  }
}

// GpBitBlt16 is quicker, but needs 8 bytes boundaries
int BitBlt16(GPDRAWTAG *gptag, GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height,
             unsigned char *bitmap, int bitmapx, int bitmapy,
             int bitmapWidth, int bitmapHeight)
{
  GPDRAWTAG limits;
  int i, j;
  unsigned short *bmp, *screen;

  // Clip bitmap
  if (gptag)
    limits = *gptag;
  else
  {
    limits.clip_x = 0;
    limits.clip_y = 0;
    limits.clip_w = 320;
    limits.clip_h = 240;
  }

  //- If completly out of limits, return
  if ((dx >= limits.clip_x + limits.clip_w) || (dy >= limits.clip_y + limits.clip_h))
    return 0;
  if ((dx + width - 1 < limits.clip_x) || (dy + height - 1 < limits.clip_y))
    return 0;

  //- Clip X and Width
  if (dx < limits.clip_x)
  {
    width -= limits.clip_x - dx;
    bitmapx += limits.clip_x - dx;
    dx = limits.clip_x;
  }
  if (dx + width > limits.clip_x + limits.clip_w)
    width = limits.clip_x + limits.clip_w - dx;

  //- Clip Y and Height
  if (dy < limits.clip_y)
  {
    height -= limits.clip_y - dy;
    bitmapy += limits.clip_y - dy;
    dy = limits.clip_y;
  }
  if (dy + height> limits.clip_y + limits.clip_h)
    height = limits.clip_y + limits.clip_h - dy;


  // Adjust to rotated screen
  //- Source coordinates
  i = bitmapx;
  bitmapx = bitmapHeight - (bitmapy + height);
  bitmapy = i;
  //- Destination coordinates
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  //- Sizes
  bitmapWidth = bitmapHeight;
  i = width;
  width = height;
  height = i;


  // Blit bitmap
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  bmp = (unsigned short *) bitmap + bitmapy * bitmapWidth + bitmapx;
  for (j=0; j < height; j++)
  {
    gm_memcpy(screen, bmp, width * sizeof(unsigned short));
    screen += width;
    bmp += width;
  }

  return 0;
}

int BitBltGrayTo16(GPDRAWTAG *gptag, GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height,
                   unsigned char *bitmap, int bitmapx, int bitmapy, int bitmapWidth, int bitmapHeight)
{
  GPDRAWTAG limits;
  int i, j;
  unsigned char *bmp;
  unsigned short *screen, color;

  // Clip bitmap
  if (gptag)
    limits = *gptag;
  else
  {
    limits.clip_x = 0;
    limits.clip_y = 0;
    limits.clip_w = 320;
    limits.clip_h = 240;
  }

  //- If completly out of limits, return
  if ((dx >= limits.clip_x + limits.clip_w) || (dy >= limits.clip_y + limits.clip_h))
    return 0;
  if ((dx + width - 1 < limits.clip_x) || (dy + height - 1 < limits.clip_y))
    return 0;

  //- Clip X and Width
  if (dx < limits.clip_x)
  {
    width -= limits.clip_x - dx;
    bitmapx += limits.clip_x - dx;
    dx = limits.clip_x;
  }
  if (dx + width > limits.clip_x + limits.clip_w)
    width = limits.clip_x + limits.clip_w - dx;

  //- Clip Y and Height
  if (dy < limits.clip_y)
  {
    height -= limits.clip_y - dy;
    bitmapy += limits.clip_y - dy;
    dy = limits.clip_y;
  }
  if (dy + height> limits.clip_y + limits.clip_h)
    height = limits.clip_y + limits.clip_h - dy;


  // Adjust to rotated screen
  //- Source coordinates
  i = bitmapx;
  bitmapx = bitmapHeight - (bitmapy + height);
  bitmapy = i;
  //- Destination coordinates
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  //- Sizes
  bitmapWidth = bitmapHeight;
  i = width;
  width = height;
  height = i;


  // Blit bitmap
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  bmp = bitmap + bitmapy * bitmapWidth + bitmapx;
  for (j=0; j < height; j++)
  {
    for (i=0; i < width; i++, screen++, bmp++)
    {
      color = *bmp >> 3;
      *screen = (color << 11) | (color << 6) | (color << 1);
    }
    screen += GPC_LCD_PHYSICAL_W - width;
    bmp += bitmapWidth - width;
  }

  return 0;
}

int BitTile16(GPDRAWTAG *gptag, GPDRAWSURFACE *ptgpds,
              int dx, int dy, int width, int height,
              unsigned char *bitmap, int bitmapx, int bitmapy,
              int bitmapWidth, int bitmapHeight, int tileWidth, int tileHeight)
{
  int x, y;
  int w, h;
  
  for (y = 0; y < height; y += tileHeight)
    for (x = 0; x < width; x += tileWidth)
    {
      w = (x + tileWidth > width) ? width - x : tileWidth;
      h = (y + tileHeight > height) ? height - y : tileHeight;
      GpBitBlt16(gptag, ptgpds, dx + x, dy + y, w, h, 
                 bitmap, bitmapx, bitmapy, bitmapWidth, bitmapHeight);
    }

  return 0;
}

int InitializeAlphaLUT(int alphaLevels, unsigned char (*pAlphaLUT)[32][32])
{
  int bottom, top, level, max;

  // If LUT is provided, use it
  if (pAlphaLUT != NULL)
  {
    giAlphaLevels = alphaLevels;
    gtAlphaLUT = pAlphaLUT;
  }

  // If not, create a new LUT
  else
  {
    giAlphaLevels = alphaLevels;
    gtAlphaLUT = gm_malloc((32 >> alphaLevels) * 32 * 32 * sizeof(unsigned char));
    if (gtAlphaLUT == NULL)
      return -1;
  }

  // Generate values
  max = (32 >> alphaLevels) - 1;
  for (level=0; level < 32 >> alphaLevels; level++)
    for (bottom=0; bottom < 32; bottom++)
      for (top=0; top < 32; top++)
        gtAlphaLUT[level][bottom][top] = ((top * level) + (bottom * (max - level))) / max;

  return 0;
}

void OverlayColor16(GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height,
                    int color, int intensity)
{
  int i, j;
  unsigned short *screen;
  int red, green, blue;
  unsigned char *pLUT;

  // Adjust to rotated screen
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  i = width;
  width = height;
  height = i;

  // Overlay color
  pLUT = &gtAlphaLUT[intensity >> giAlphaLevels][0][0];
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  for (j=0; j < height; j++)
  {
    for (i=0; i < width; i++, screen++)
    {
      red   = pLUT[GetRGBRed(*screen)   * 32 + GetRGBRed  (color)];
      green = pLUT[GetRGBGreen(*screen) * 32 + GetRGBGreen(color)];
      blue  = pLUT[GetRGBBlue(*screen)  * 32 + GetRGBBlue (color)];

      *screen = RGB(red, green, blue);
    }

    screen += GPC_LCD_PHYSICAL_W - width;
  }
}

void ReleaseAlphaLUT(void)
{
  gm_free(gtAlphaLUT);
}

void ReleaseScreen(unsigned short *bitmap)
{
  // releases memory
  if (bitmap != NULL)
    gm_free(bitmap);
}

void ReleaseScreenRegion(unsigned short *bitmap)
{
  // releases memory
  if (bitmap != NULL)
    gm_free(bitmap);
}

unsigned short *SaveBackScreen(void)
{
  unsigned short *bitmap;
  
  //-- save current back screen state
  bitmap = (unsigned short *) gm_malloc(320 * 240 * sizeof(unsigned short));
  if (bitmap != NULL)
    gm_memcpy(bitmap, gtSurface[giSurface].ptbuffer, 320 * 240 * sizeof(unsigned short));

  return bitmap;
}

unsigned short *SaveBackScreenRegion(int x, int y, int width, int height)
{
  unsigned short *bitmap;
  int i, j;

  //-- save back screen region
  bitmap = (unsigned short *) gm_malloc(width * height * sizeof(unsigned short));
  if (bitmap != NULL)
    for (j=0; j < height; j++)
      for (i=0; i < width; i++)
        bitmap[i * height + ((height - 1) - j)] = GetPixel(&gtSurface[giSurface], x + i, y + j);
//        bitmap[i * 240 + (239 - j)] = ((unsigned short *)gtSurface[giSurface].ptbuffer)[i * 240 + (239 - j)];

  return bitmap;
}

unsigned short *SaveScreen(void)
{
  unsigned short *screen;
  
  //-- save current screen state
  screen = (unsigned short *) gm_malloc(320 * 240 * sizeof(unsigned short));
  if (screen != NULL)
    gm_memcpy(screen, gtSurface[(giSurface+1) & 0x01].ptbuffer, 320 * 240 * sizeof(unsigned short));

  return screen;
}

void SubtractColor16(GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height, int color)
{
  int i, j;
  unsigned short *screen;
  int red, green, blue;

  // Adjust to rotated screen
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  i = width;
  width = height;
  height = i;


  // Subtract color
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  for (j=0; j < height; j++)
  {
    for (i=0; i < width; i++, screen++)
    {
      red   = GetRGBRed  (*screen) - GetRGBRed  (color);
      green = GetRGBGreen(*screen) - GetRGBGreen(color);
      blue  = GetRGBBlue (*screen) - GetRGBBlue (color);
      if (red < 0)   red = 0;
      if (green < 0) green = 0;
      if (blue < 0)  blue = 0;

      *screen = RGB(red, green, blue);
    }

    screen += GPC_LCD_PHYSICAL_W - width;
  }
}

// For small bitmaps GpTransBlt16 is quicker, but needs 8 bytes boundaries
int TransBlt16(GPDRAWTAG *gptag, GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height,
               unsigned char *bitmap, int bitmapx, int bitmapy,
               int bitmapWidth, int bitmapHeight, unsigned short transColor)
{
  GPDRAWTAG limits;
  int i, j;
  unsigned short *bmp, *screen;

  // Clip bitmap
  if (gptag)
    limits = *gptag;
  else
  {
    limits.clip_x = 0;
    limits.clip_y = 0;
    limits.clip_w = 320;
    limits.clip_h = 240;
  }

  //- If completly out of limits, return
  if ((dx >= limits.clip_x + limits.clip_w) || (dy >= limits.clip_y + limits.clip_h))
    return 0;
  if ((dx + width - 1 < limits.clip_x) || (dy + height - 1 < limits.clip_y))
    return 0;

  //- Clip X and Width
  if (dx < limits.clip_x)
  {
    width -= limits.clip_x - dx;
    bitmapx += limits.clip_x - dx;
    dx = limits.clip_x;
  }
  if (dx + width > limits.clip_x + limits.clip_w)
    width = limits.clip_x + limits.clip_w - dx;

  //- Clip Y and Height
  if (dy < limits.clip_y)
  {
    height -= limits.clip_y - dy;
    bitmapy += limits.clip_y - dy;
    dy = limits.clip_y;
  }
  if (dy + height> limits.clip_y + limits.clip_h)
    height = limits.clip_y + limits.clip_h - dy;


  // Adjust to rotated screen
  //- Source coordinates
  i = bitmapx;
  bitmapx = bitmapHeight - (bitmapy + height);
  bitmapy = i;
  //- Destination coordinates
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  //- Sizes
  bitmapWidth = bitmapHeight;
  i = width;
  width = height;
  height = i;

  
  // Blit bitmap
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  bmp = (unsigned short *) bitmap + bitmapy * bitmapWidth + bitmapx;
  for (j=0; j < height; j++)
  {
    for (i=0; i < width; i++, screen++, bmp++)
      if (*bmp != transColor)
        *screen = *bmp;
    screen += GPC_LCD_PHYSICAL_W - width;
    bmp += bitmapWidth - width;
  }

  return 0;
}

int TransFade16(GPDRAWTAG *gptag, GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height,
                unsigned char *bitmap, int bitmapx, int bitmapy,
                int bitmapWidth, int bitmapHeight, unsigned short transColor, int opacity)
{
  GPDRAWTAG limits;
  int i, j;
  unsigned short *bmp, *screen;
  int red, green, blue;
  unsigned char *pLUT;

  if (opacity == 0)
    return 0;


  // Clip bitmap
  if (gptag)
    limits = *gptag;
  else
  {
    limits.clip_x = 0;
    limits.clip_y = 0;
    limits.clip_w = 320;
    limits.clip_h = 240;
  }

  //- If completly out of limits, return
  if ((dx >= limits.clip_x + limits.clip_w) || (dy >= limits.clip_y + limits.clip_h))
    return 0;
  if ((dx + width - 1 < limits.clip_x) || (dy + height - 1 < limits.clip_y))
    return 0;

  //- Clip X and Width
  if (dx < limits.clip_x)
  {
    width -= limits.clip_x - dx;
    bitmapx += limits.clip_x - dx;
    dx = limits.clip_x;
  }
  if (dx + width > limits.clip_x + limits.clip_w)
    width = limits.clip_x + limits.clip_w - dx;

  //- Clip Y and Height
  if (dy < limits.clip_y)
  {
    height -= limits.clip_y - dy;
    bitmapy += limits.clip_y - dy;
    dy = limits.clip_y;
  }
  if (dy + height> limits.clip_y + limits.clip_h)
    height = limits.clip_y + limits.clip_h - dy;


  // Adjust to rotated screen
  //- Source coordinates
  i = bitmapx;
  bitmapx = bitmapHeight - (bitmapy + height);
  bitmapy = i;
  //- Destination coordinates
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  //- Sizes
  bitmapWidth = bitmapHeight;
  i = width;
  width = height;
  height = i;


  // Fade bitmap
  pLUT = &gtAlphaLUT[opacity >> giAlphaLevels][0][0];
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  bmp = (unsigned short *) bitmap + bitmapy * bitmapWidth + bitmapx;
  for (j=0; j < height; j++)
  {
    for (i=0; i < width; i++, screen++, bmp++)
      // If not transparent color, process
      if (*bmp != transColor)
      {
        // Merge color with background
        red   = pLUT[GetRGBRed(*screen)   * 32 + GetRGBRed(*bmp)  ];
        green = pLUT[GetRGBGreen(*screen) * 32 + GetRGBGreen(*bmp)];
        blue  = pLUT[GetRGBBlue(*screen)  * 32 + GetRGBBlue(*bmp) ];

        // Put pixel
        *screen = RGB(red, green, blue);
      }

    // Jump to next scan line
    screen += GPC_LCD_PHYSICAL_W - width;
    bmp += bitmapWidth - width;
  }

  return 0;
}

int TransFadeMask16(GPDRAWTAG *gptag, GPDRAWSURFACE *ptgpds, int dx, int dy, int width, int height,
                    unsigned char *mask, int maskx, int masky,
                    int maskWidth, int maskHeight, unsigned short color, int opacity)
{
  GPDRAWTAG limits;
  int i, j;
  unsigned short *msk, *screen;
  unsigned short level;
  int red, green, blue;
  int colorRed, colorGreen, colorBlue;

  if (opacity == 0)
    return 0;


  // Clip mask
  if (gptag)
    limits = *gptag;
  else
  {
    limits.clip_x = 0;
    limits.clip_y = 0;
    limits.clip_w = 320;
    limits.clip_h = 240;
  }

  //- If completly out of limits, return
  if ((dx >= limits.clip_x + limits.clip_w) || (dy >= limits.clip_y + limits.clip_h))
    return 0;
  if ((dx + width - 1 < limits.clip_x) || (dy + height - 1 < limits.clip_y))
    return 0;

  //- Clip X and Width
  if (dx < limits.clip_x)
  {
    width -= limits.clip_x - dx;
    maskx += limits.clip_x - dx;
    dx = limits.clip_x;
  }
  if (dx + width > limits.clip_x + limits.clip_w)
    width = limits.clip_x + limits.clip_w - dx;

  //- Clip Y and Height
  if (dy < limits.clip_y)
  {
    height -= limits.clip_y - dy;
    masky += limits.clip_y - dy;
    dy = limits.clip_y;
  }
  if (dy + height> limits.clip_y + limits.clip_h)
    height = limits.clip_y + limits.clip_h - dy;


  // Adjust to rotated screen
  //- Source coordinates
  i = maskx;
  maskx = maskHeight - (masky + height);
  masky = i;
  //- Destination coordinates
  i = dx;
  dx = GPC_LCD_HEIGHT - (dy + height);
  dy = i;
  //- Sizes
  maskWidth = maskHeight;
  i = width;
  width = height;
  height = i;


  // Separate color components
  colorRed   = GetRGBRed(color);
  colorGreen = GetRGBGreen(color);
  colorBlue  = GetRGBBlue(color);

  
  // Draw bitmap using mask & color & opacity
  screen = (unsigned short *) (ptgpds->ptbuffer) + dy * GPC_LCD_PHYSICAL_W + dx;
  msk = (unsigned short *) mask + masky * maskWidth + maskx;
  for (j=0; j < height; j++)
  {
    for (i=0; i < width; i++, screen++, msk++)
    {
      // Get mask pixel level
      level = (*msk >> 1) & 0x1F;

      if (level > 0)
      {
        // Adjust level of opacity
        if (opacity < 31)
          level = (level * opacity) / 31;
          
        if (level == 31)
        {
            // Put pixel, as is
            *screen = color;
        }
        else
          if (level > 0)
          {
            // Merge color with background
            red   = gtAlphaLUT[level >> giAlphaLevels] [GetRGBRed(*screen)  ] [colorRed  ];
            green = gtAlphaLUT[level >> giAlphaLevels] [GetRGBGreen(*screen)] [colorGreen];
            blue  = gtAlphaLUT[level >> giAlphaLevels] [GetRGBBlue(*screen) ] [colorBlue ];

            // Put pixel
            *screen = RGB(red, green, blue);
          }
      }
    }

    // Jump to next scan line
    screen += GPC_LCD_PHYSICAL_W - width;
    msk += maskWidth - width;
  }

  return 0;
}

int TransTile16(GPDRAWTAG *gptag, GPDRAWSURFACE *ptgpds,
                int dx, int dy, int width, int height,
                unsigned char *bitmap, int bitmapx, int bitmapy,
                int bitmapWidth, int bitmapHeight, int tileWidth, int tileHeight, int transColor)
{
  int x, y;
  int w, h;
  
  for (y = 0; y < height; y += tileHeight)
    for (x = 0; x < width; x += tileWidth)
    {
      w = (x + tileWidth > width) ? width - x : tileWidth;
      h = (y + tileHeight > height) ? height - y : tileHeight;
      GpTransBlt16(gptag, ptgpds, dx + x, dy + y, w, h, 
                   bitmap, bitmapx, bitmapy, bitmapWidth, bitmapHeight, transColor);
    }

  return 0;
}
